# DO NOT EDIT -- GENERATED BY: `generate-tool.py --generate`
__all__ = ("SETTINGS_TOPOLOGICALLY_SORTED", "_Setting")

import sys
from typing import Tuple

if sys.version_info >= (3, 8):
    from typing import Final, Literal
else:
    from typing_extensions import Final, Literal


_Setting = Literal[
    "_args",
    "_aws_lambda",
    "_cli_only_mode",
    "_code_path_local",
    "_colab",
    "_cuda",
    "_disable_meta",
    "_disable_service",
    "_disable_setproctitle",
    "_disable_stats",
    "_disable_update_check",
    "_disable_viewer",
    "_disable_machine_info",
    "_executable",
    "_extra_http_headers",
    "_file_stream_max_bytes",
    "_file_stream_retry_max",
    "_file_stream_retry_wait_min_seconds",
    "_file_stream_retry_wait_max_seconds",
    "_file_stream_timeout_seconds",
    "_file_transfer_retry_max",
    "_file_transfer_retry_wait_min_seconds",
    "_file_transfer_retry_wait_max_seconds",
    "_file_transfer_timeout_seconds",
    "_flow_control_custom",
    "_flow_control_disabled",
    "_graphql_retry_max",
    "_graphql_retry_wait_min_seconds",
    "_graphql_retry_wait_max_seconds",
    "_graphql_timeout_seconds",
    "_internal_check_process",
    "_internal_queue_timeout",
    "_ipython",
    "_jupyter",
    "_jupyter_name",
    "_jupyter_path",
    "_jupyter_root",
    "_kaggle",
    "_live_policy_rate_limit",
    "_live_policy_wait_time",
    "_log_level",
    "_network_buffer",
    "_noop",
    "_notebook",
    "_offline",
    "_sync",
    "_os",
    "_platform",
    "_proxies",
    "_python",
    "_runqueue_item_id",
    "_require_legacy_service",
    "_save_requirements",
    "_service_transport",
    "_service_wait",
    "_shared",
    "_start_datetime",
    "_start_time",
    "_stats_pid",
    "_stats_sampling_interval",
    "_stats_sample_rate_seconds",
    "_stats_samples_to_average",
    "_stats_join_assets",
    "_stats_neuron_monitor_config_path",
    "_stats_open_metrics_endpoints",
    "_stats_open_metrics_filters",
    "_stats_disk_paths",
    "_stats_buffer_size",
    "_tmp_code_dir",
    "_tracelog",
    "_unsaved_keys",
    "_windows",
    "allow_val_change",
    "anonymous",
    "api_key",
    "azure_account_url_to_access_key",
    "base_url",
    "code_dir",
    "colab_url",
    "config_paths",
    "console",
    "console_multipart",
    "credentials_file",
    "deployment",
    "disable_code",
    "disable_git",
    "disable_hints",
    "disable_job_creation",
    "disabled",
    "docker",
    "email",
    "entity",
    "files_dir",
    "force",
    "fork_from",
    "resume_from",
    "git_commit",
    "git_remote",
    "git_remote_url",
    "git_root",
    "heartbeat_seconds",
    "host",
    "http_proxy",
    "https_proxy",
    "identity_token_file",
    "ignore_globs",
    "init_timeout",
    "is_local",
    "job_name",
    "job_source",
    "label_disable",
    "launch",
    "launch_config_path",
    "log_dir",
    "log_internal",
    "log_symlink_internal",
    "log_symlink_user",
    "log_user",
    "login_timeout",
    "mode",
    "notebook_name",
    "program",
    "program_abspath",
    "program_relpath",
    "project",
    "project_url",
    "quiet",
    "reinit",
    "relogin",
    "resume",
    "resume_fname",
    "resumed",
    "root_dir",
    "run_group",
    "run_id",
    "run_job_type",
    "run_mode",
    "run_name",
    "run_notes",
    "run_tags",
    "run_url",
    "sagemaker_disable",
    "save_code",
    "settings_system",
    "settings_workspace",
    "show_colors",
    "show_emoji",
    "show_errors",
    "show_info",
    "show_warnings",
    "silent",
    "start_method",
    "strict",
    "summary_errors",
    "summary_timeout",
    "summary_warnings",
    "sweep_id",
    "sweep_param_path",
    "sweep_url",
    "symlink",
    "sync_dir",
    "sync_file",
    "sync_symlink_latest",
    "table_raise_on_max_row_limit_exceeded",
    "timespec",
    "tmp_dir",
    "username",
    "wandb_dir",
]

SETTINGS_TOPOLOGICALLY_SORTED: Final[Tuple[_Setting, ...]] = (
    "_service_wait",
    "_stats_sample_rate_seconds",
    "_stats_samples_to_average",
    "_stats_sampling_interval",
    "anonymous",
    "api_key",
    "base_url",
    "console",
    "job_source",
    "mode",
    "project",
    "run_id",
    "start_method",
    "_aws_lambda",
    "program",
    "_code_path_local",
    "_colab",
    "_disable_machine_info",
    "_disable_meta",
    "_disable_stats",
    "_network_buffer",
    "_flow_control_disabled",
    "_flow_control_custom",
    "_ipython",
    "_jupyter",
    "_kaggle",
    "_noop",
    "_notebook",
    "disabled",
    "_offline",
    "_shared",
    "_stats_neuron_monitor_config_path",
    "run_mode",
    "_start_datetime",
    "timespec",
    "root_dir",
    "wandb_dir",
    "tmp_dir",
    "_tmp_code_dir",
    "_windows",
    "colab_url",
    "is_local",
    "deployment",
    "disable_code",
    "disable_git",
    "disable_job_creation",
    "files_dir",
    "_proxies",
    "http_proxy",
    "https_proxy",
    "log_dir",
    "log_internal",
    "log_symlink_internal",
    "log_symlink_user",
    "log_user",
    "project_url",
    "resume_fname",
    "run_url",
    "settings_system",
    "settings_workspace",
    "sweep_url",
    "sync_dir",
    "sync_file",
    "sync_symlink_latest",
)
