mutation CreateRunFiles($entity: String!, $project: String!, $run: String!, $files: [String!]!) {
    createRunFiles(input: {entityName: $entity, projectName: $project, runName: $run, files: $files}) {
        runID
        uploadHeaders
        files {
            name
            uploadUrl
        }
    }
}
