# Description: This file contains the query to fetch the status of a run resume.

query RunResumeStatus($project: String, $entity: String, $name: String!) {
  model(name: $project, entityName: $entity) {
    id
    name
    entity {
      id
      name
    }

    bucket(name: $name, missingOk: true) {
      id
      name
      summaryMetrics
      displayName
      logLineCount
      historyLineCount
      eventsLineCount
      historyTail
      eventsTail
      config
      tags
      # Pulling wandbConfig.t is required so that runresume.go can determine if a run has actually started
      wandbConfig(keys: ["t"])
    }
  }
}
