# @genqlient(omitempty: true)
query Viewer{
    viewer {
        id
        entity
        flags
        teams {
            edges {
                node {
                    name
                }
            }
        }
    }
}
