# DO NOT EDIT -- GENERATED BY: `generate-tool.py --generate`
import sys

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


URLS = Literal[
    "cli_launch",
    "doc_run",
    "doc_require",
    "doc_start_err",
    "doc_artifacts_guide",
    "upgrade_server",
    "multiprocess",
    "wandb_init",
    "wandb_server",
    "wandb_define_metric",
    "wandb_core",
]
