// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.33.0
// 	protoc        v4.23.4
// source: wandb/proto/wandb_base.proto

package service_go_proto

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// _RecordInfo, _RequestInfo: extra info for all records and requests
type XRecordInfo struct {
	state         protoimpl.MessageState
	sizeCache     protoimpl.SizeCache
	unknownFields protoimpl.UnknownFields

	StreamId    string `protobuf:"bytes,1,opt,name=stream_id,json=streamId,proto3" json:"stream_id,omitempty"`
	XTracelogId string `protobuf:"bytes,100,opt,name=_tracelog_id,json=TracelogId,proto3" json:"_tracelog_id,omitempty"`
}

func (x *XRecordInfo) Reset() {
	*x = XRecordInfo{}
	if protoimpl.UnsafeEnabled {
		mi := &file_wandb_proto_wandb_base_proto_msgTypes[0]
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		ms.StoreMessageInfo(mi)
	}
}

func (x *XRecordInfo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*XRecordInfo) ProtoMessage() {}

func (x *XRecordInfo) ProtoReflect() protoreflect.Message {
	mi := &file_wandb_proto_wandb_base_proto_msgTypes[0]
	if protoimpl.UnsafeEnabled && x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use XRecordInfo.ProtoReflect.Descriptor instead.
func (*XRecordInfo) Descriptor() ([]byte, []int) {
	return file_wandb_proto_wandb_base_proto_rawDescGZIP(), []int{0}
}

func (x *XRecordInfo) GetStreamId() string {
	if x != nil {
		return x.StreamId
	}
	return ""
}

func (x *XRecordInfo) GetXTracelogId() string {
	if x != nil {
		return x.XTracelogId
	}
	return ""
}

type XRequestInfo struct {
	state         protoimpl.MessageState
	sizeCache     protoimpl.SizeCache
	unknownFields protoimpl.UnknownFields

	StreamId string `protobuf:"bytes,1,opt,name=stream_id,json=streamId,proto3" json:"stream_id,omitempty"`
}

func (x *XRequestInfo) Reset() {
	*x = XRequestInfo{}
	if protoimpl.UnsafeEnabled {
		mi := &file_wandb_proto_wandb_base_proto_msgTypes[1]
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		ms.StoreMessageInfo(mi)
	}
}

func (x *XRequestInfo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*XRequestInfo) ProtoMessage() {}

func (x *XRequestInfo) ProtoReflect() protoreflect.Message {
	mi := &file_wandb_proto_wandb_base_proto_msgTypes[1]
	if protoimpl.UnsafeEnabled && x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use XRequestInfo.ProtoReflect.Descriptor instead.
func (*XRequestInfo) Descriptor() ([]byte, []int) {
	return file_wandb_proto_wandb_base_proto_rawDescGZIP(), []int{1}
}

func (x *XRequestInfo) GetStreamId() string {
	if x != nil {
		return x.StreamId
	}
	return ""
}

// _ResultInfo: extra info for all results
type XResultInfo struct {
	state         protoimpl.MessageState
	sizeCache     protoimpl.SizeCache
	unknownFields protoimpl.UnknownFields

	XTracelogId string `protobuf:"bytes,100,opt,name=_tracelog_id,json=TracelogId,proto3" json:"_tracelog_id,omitempty"`
}

func (x *XResultInfo) Reset() {
	*x = XResultInfo{}
	if protoimpl.UnsafeEnabled {
		mi := &file_wandb_proto_wandb_base_proto_msgTypes[2]
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		ms.StoreMessageInfo(mi)
	}
}

func (x *XResultInfo) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*XResultInfo) ProtoMessage() {}

func (x *XResultInfo) ProtoReflect() protoreflect.Message {
	mi := &file_wandb_proto_wandb_base_proto_msgTypes[2]
	if protoimpl.UnsafeEnabled && x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use XResultInfo.ProtoReflect.Descriptor instead.
func (*XResultInfo) Descriptor() ([]byte, []int) {
	return file_wandb_proto_wandb_base_proto_rawDescGZIP(), []int{2}
}

func (x *XResultInfo) GetXTracelogId() string {
	if x != nil {
		return x.XTracelogId
	}
	return ""
}

var File_wandb_proto_wandb_base_proto protoreflect.FileDescriptor

var file_wandb_proto_wandb_base_proto_rawDesc = []byte{
	0x0a, 0x1c, 0x77, 0x61, 0x6e, 0x64, 0x62, 0x2f, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x2f, 0x77, 0x61,
	0x6e, 0x64, 0x62, 0x5f, 0x62, 0x61, 0x73, 0x65, 0x2e, 0x70, 0x72, 0x6f, 0x74, 0x6f, 0x12, 0x0e,
	0x77, 0x61, 0x6e, 0x64, 0x62, 0x5f, 0x69, 0x6e, 0x74, 0x65, 0x72, 0x6e, 0x61, 0x6c, 0x22, 0x4c,
	0x0a, 0x0b, 0x5f, 0x52, 0x65, 0x63, 0x6f, 0x72, 0x64, 0x49, 0x6e, 0x66, 0x6f, 0x12, 0x1b, 0x0a,
	0x09, 0x73, 0x74, 0x72, 0x65, 0x61, 0x6d, 0x5f, 0x69, 0x64, 0x18, 0x01, 0x20, 0x01, 0x28, 0x09,
	0x52, 0x08, 0x73, 0x74, 0x72, 0x65, 0x61, 0x6d, 0x49, 0x64, 0x12, 0x20, 0x0a, 0x0c, 0x5f, 0x74,
	0x72, 0x61, 0x63, 0x65, 0x6c, 0x6f, 0x67, 0x5f, 0x69, 0x64, 0x18, 0x64, 0x20, 0x01, 0x28, 0x09,
	0x52, 0x0a, 0x54, 0x72, 0x61, 0x63, 0x65, 0x6c, 0x6f, 0x67, 0x49, 0x64, 0x22, 0x2b, 0x0a, 0x0c,
	0x5f, 0x52, 0x65, 0x71, 0x75, 0x65, 0x73, 0x74, 0x49, 0x6e, 0x66, 0x6f, 0x12, 0x1b, 0x0a, 0x09,
	0x73, 0x74, 0x72, 0x65, 0x61, 0x6d, 0x5f, 0x69, 0x64, 0x18, 0x01, 0x20, 0x01, 0x28, 0x09, 0x52,
	0x08, 0x73, 0x74, 0x72, 0x65, 0x61, 0x6d, 0x49, 0x64, 0x22, 0x2f, 0x0a, 0x0b, 0x5f, 0x52, 0x65,
	0x73, 0x75, 0x6c, 0x74, 0x49, 0x6e, 0x66, 0x6f, 0x12, 0x20, 0x0a, 0x0c, 0x5f, 0x74, 0x72, 0x61,
	0x63, 0x65, 0x6c, 0x6f, 0x67, 0x5f, 0x69, 0x64, 0x18, 0x64, 0x20, 0x01, 0x28, 0x09, 0x52, 0x0a,
	0x54, 0x72, 0x61, 0x63, 0x65, 0x6c, 0x6f, 0x67, 0x49, 0x64, 0x62, 0x06, 0x70, 0x72, 0x6f, 0x74,
	0x6f, 0x33,
}

var (
	file_wandb_proto_wandb_base_proto_rawDescOnce sync.Once
	file_wandb_proto_wandb_base_proto_rawDescData = file_wandb_proto_wandb_base_proto_rawDesc
)

func file_wandb_proto_wandb_base_proto_rawDescGZIP() []byte {
	file_wandb_proto_wandb_base_proto_rawDescOnce.Do(func() {
		file_wandb_proto_wandb_base_proto_rawDescData = protoimpl.X.CompressGZIP(file_wandb_proto_wandb_base_proto_rawDescData)
	})
	return file_wandb_proto_wandb_base_proto_rawDescData
}

var file_wandb_proto_wandb_base_proto_msgTypes = make([]protoimpl.MessageInfo, 3)
var file_wandb_proto_wandb_base_proto_goTypes = []interface{}{
	(*XRecordInfo)(nil),  // 0: wandb_internal._RecordInfo
	(*XRequestInfo)(nil), // 1: wandb_internal._RequestInfo
	(*XResultInfo)(nil),  // 2: wandb_internal._ResultInfo
}
var file_wandb_proto_wandb_base_proto_depIdxs = []int32{
	0, // [0:0] is the sub-list for method output_type
	0, // [0:0] is the sub-list for method input_type
	0, // [0:0] is the sub-list for extension type_name
	0, // [0:0] is the sub-list for extension extendee
	0, // [0:0] is the sub-list for field type_name
}

func init() { file_wandb_proto_wandb_base_proto_init() }
func file_wandb_proto_wandb_base_proto_init() {
	if File_wandb_proto_wandb_base_proto != nil {
		return
	}
	if !protoimpl.UnsafeEnabled {
		file_wandb_proto_wandb_base_proto_msgTypes[0].Exporter = func(v interface{}, i int) interface{} {
			switch v := v.(*XRecordInfo); i {
			case 0:
				return &v.state
			case 1:
				return &v.sizeCache
			case 2:
				return &v.unknownFields
			default:
				return nil
			}
		}
		file_wandb_proto_wandb_base_proto_msgTypes[1].Exporter = func(v interface{}, i int) interface{} {
			switch v := v.(*XRequestInfo); i {
			case 0:
				return &v.state
			case 1:
				return &v.sizeCache
			case 2:
				return &v.unknownFields
			default:
				return nil
			}
		}
		file_wandb_proto_wandb_base_proto_msgTypes[2].Exporter = func(v interface{}, i int) interface{} {
			switch v := v.(*XResultInfo); i {
			case 0:
				return &v.state
			case 1:
				return &v.sizeCache
			case 2:
				return &v.unknownFields
			default:
				return nil
			}
		}
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: file_wandb_proto_wandb_base_proto_rawDesc,
			NumEnums:      0,
			NumMessages:   3,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_wandb_proto_wandb_base_proto_goTypes,
		DependencyIndexes: file_wandb_proto_wandb_base_proto_depIdxs,
		MessageInfos:      file_wandb_proto_wandb_base_proto_msgTypes,
	}.Build()
	File_wandb_proto_wandb_base_proto = out.File
	file_wandb_proto_wandb_base_proto_rawDesc = nil
	file_wandb_proto_wandb_base_proto_goTypes = nil
	file_wandb_proto_wandb_base_proto_depIdxs = nil
}
