# @genqlient(omitempty: true)
query ServerInfo{
    serverInfo {
        cliVersionInfo
        latestLocalVersionInfo {
            outOfDate
            latestVersionString
            versionOnThisInstanceString
        }
    }
}
