// Package wbvalue allows saving complex run objects.
//
// This package implements data types that have nontrivial representations
// (such as charts) or complex requirements (such as tables) for uploading
// to a run.
package wbvalue
